#coding: utf-8
""" Utility programs for environment variables."""
import os
import re
import warnings
from dotenv import load_dotenv

from ._path import DOTENV_PATH
from ._warnings import EnvVariableNotDefinedWarning
from .coloring_utils  import toBLUE, toGREEN

TRANSLATION_GUMMY_ENVNAME_PREFIX = "TRANSLATION_GUMMY"

def name2envname(name, service="", prefix=""):
    """Convert name to environment varname.

    Args:
        name (str)    : name.
        service (str) : service name. (default= ``""``)
        prefix (str)  : prefix of the ``name``. (default= ``""``)

    Examples:
        >>> from gummy.utils import name2envname
        >>> name2envname(name="name", service="",        prefix="")
        'TRANSLATION_GUMMY_NAME'
        >>> name2envname(name="name", service="service", prefix="")
        'TRANSLATION_GUMMY_SERVICE_NAME'
        >>> name2envname(name="name", service="",        prefix="prefix")
        'TRANSLATION_GUMMY_PREFIX_NAME'
        >>> name2envname(name="name", service="service", prefix="prefix")
        'TRANSLATION_GUMMY_SERVICE_PREFIX_NAME'
    """
    strings = [TRANSLATION_GUMMY_ENVNAME_PREFIX]
    for string in [service, prefix]:
        if len(string)>0:
            strings.append(string)
    strings.append(name)
    return "_".join(strings).upper()

def where_is_envfile():
    """Get Where the envfile is.

    Examples:
        >>> from gummy.utils import where_is_envfile
        >>> where_is_envfile()
        '/Users/iwasakishuto/.gummy/.env'
    """
    return DOTENV_PATH

def read_environ(dotenv_path=DOTENV_PATH):
    """Read the environment variables from ``dotenv_path`` 
    
    Args:
        dotenv_path (str) : path/to/.env (default= ``DOTENV_PATH``)

    Returns:
        dict : the dictionary containing ``dotenv_path``'s environment variables.
    
    Examples:
        >>> from gummy.utils import read_environ
        >>> read_environ()
        {'TRANSLATION_GUMMY_GATEWAY_UTOKYO_USERNAME': 'USERNAME',
        'TRANSLATION_GUMMY_GATEWAY_UTOKYO_PASSWORD': 'PASSWORD'}
    """
    env_variables = {}
    if os.path.exists(dotenv_path):
        with open(dotenv_path, mode="r", encoding='utf-8') as f:
            for line in f.readlines():
                for key,val in re.findall(pattern=r'^(.+?)\s?=\s?"?(.+?)"?$', string=line):
                    env_variables[key] = val
    return env_variables

def write_environ(dotenv_path=DOTENV_PATH, **kwargs):
    """Overwrite or Write the environment variables in ``dotenv_path``
    
    Args:
        dotenv_path (str) : path/to/.env (default= ``DOTENV_PATH``)
        kwargs (dict)     : new environment variables

    Examples:
        >>> from gummy.utils import read_environ, write_environ
        >>> read_environ()
        {'TRANSLATION_GUMMY_GATEWAY_UTOKYO_USERNAME': 'USERNAME',
        'TRANSLATION_GUMMY_GATEWAY_UTOKYO_PASSWORD': 'PASSWORD'}
        >>> write_environ(TRANSLATION_GUMMY_SAMPLE="SAMPLE")
        >>> read_environ()
        {'TRANSLATION_GUMMY_GATEWAY_UTOKYO_USERNAME': '9890136422',
        'TRANSLATION_GUMMY_GATEWAY_UTOKYO_PASSWORD': 'Shuto070389&',
        'TRANSLATION_GUMMY_SAMPLE': 'SAMPLE'}
        >>> write_environ(TRANSLATION_GUMMY_SAMPLE="SAMPLE_AMENDED")
        >>> read_environ()
        {'TRANSLATION_GUMMY_GATEWAY_UTOKYO_USERNAME': '9890136422',
        'TRANSLATION_GUMMY_GATEWAY_UTOKYO_PASSWORD': 'Shuto070389&',
        'TRANSLATION_GUMMY_SAMPLE': 'SAMPLE_AMENDED'}
    """
    env_variables = read_environ(dotenv_path)
    env_variables.update(kwargs)
    with open(DOTENV_PATH, mode="w", encoding='utf-8') as f:
        f.writelines([f'{key} = "{val}"\n' for key,val in env_variables.items()])

def show_environ(dotenv_path=DOTENV_PATH):
    """Show environment variables written in ``dotenv_path``.

    Args:
        dotenv_path (str) : path/to/.env (default= ``DOTENV_PATH``)

    Examples:
        >>> from gummy.utils import show_environ
        >>> show_environ()
        * TRANSLATION_GUMMY_GATEWAY_UTOKYO_USERNAME : "9890136422"
        * TRANSLATION_GUMMY_GATEWAY_UTOKYO_PASSWORD : "Shuto070389&"    
    """
    env_variables = read_environ(dotenv_path=dotenv_path)
    for key,val in env_variables.items():
        print(f'* {toGREEN(key)} : "{toBLUE(val)}"')

def load_environ(dotenv_path=DOTENV_PATH, env_varnames=[], verbose=False):
    """
    Load environment variables from ``path`` file, and return whether every 
    necessary VARNAMES (``env_varnames``) are set. 

    Args:
        dotenv_path (str)   : path/to/.env (default= ``DOTENV_PATH``)
        env_varnames (list) : new environment variables
        verbose (bool)      : Whether to print message or not. (default= ``False``) 

    Returns:
        bool : Whether there are environment variables not defined in ``dotenv_path`` but in ``env_varnames``

    Examples:
        >>> from gummy.utils import load_environ, show_environ
        >>> is_ok = load_environ()
        >>> is_ok
        True
        >>> show_environ()
        * TRANSLATION_GUMMY_GATEWAY_UTOKYO_USERNAME : "9890136422"
        * TRANSLATION_GUMMY_GATEWAY_UTOKYO_PASSWORD : "Shuto070389&"
        >>> is_ok = load_environ(env_varnames=["TRANSLATION_GUMMY_NOT_DEFINED"])
        TRANSLATION_GUMMY_NOT_DEFINED is not set.
        EnvVariableNotDefinedWarning: Please set environment variable in /Users/iwasakishuto/.gummy/.env
        >>> is_ok
        False
    """
    if not os.path.exists(dotenv_path):
        return False
    load_dotenv(dotenv_path=dotenv_path, verbose=verbose)

    omission = False
    for env_name in env_varnames:
        if os.getenv(env_name) is None:
            omission = True
            print(f"{toGREEN(env_name)} is not set.")
    if omission:
        warnings.warn(message=f"Please set environment variable in {toBLUE(dotenv_path)}", category=EnvVariableNotDefinedWarning)
    return not omission

def check_environ(required_keynames, required_env_varnames=None, verbose=True, **kwargs):
    """Check whether meet the requirements.

    Args:
        required_keynames (list)     : Required keynames
        required_env_varnames (list) : Required environment variables.
        verbose (bool)               : Whether to print message or not. (default= ``True``) 

    Examples:
        >>> check_environ(
        ...     required_keynames=["hoge"],
        ...     required_env_varnames=["TRANSLATION_GUMMY_HOGE"],
        ...     hoge="hoge"
        >>> )
        (True, [])
        >>> check_environ(
        ...     required_keynames=["hoge"],
        ...     required_env_varnames=["TRANSLATION_GUMMY_HOGE"],
        ...     hoge1="hoge"
        >>> )
        Please set TRANSLATION_GUMMY_HOGE or give hoge as kwargs.
        (False, ['hoge'])

    Returns:
        tuple (bool, list) : 
            - ``is_ok`` : Whether meet the requirements.
            - ``not_meet_keynames`` : contains keynames you have to supply.
    """
    if required_env_varnames is None:
        required_env_varnames = [name2envname(name=keyname) for keyname in required_keynames]
    not_meet_keynames = []
    given_keynames = list(kwargs.keys())
    for keyname,env_name in zip(required_keynames, required_env_varnames):
        if (keyname not in given_keynames) and (os.getenv(env_name) is None):
            not_meet_keynames.append(keyname)
            if verbose: print(f"Please set {toGREEN(env_name)} or give {toBLUE(keyname)} as kwargs.")
    return len(not_meet_keynames)==0, not_meet_keynames



# coding: utf-8
from ._exceptions import *
from ._path import *
from ._warnings import *
from . import coloring_utils
from . import compress_utils
from . import download_utils
from . import driver_utils
from . import environ_utils
from . import generic_utils
from . import journal_utils
from . import monitor_utils
from . import outfmt_utils
from . import pdf_utils
from . import soup_utils

from .coloring_utils import (toGRAY, toRED, toGREEN, toYELLOW, toBLUE, toPURPLE, 
                             toCYAN, toWHITE, toREVERSE, toACCENT, toFLASH, toRED_FLASH)

from .compress_utils import recreate_dir
from .compress_utils import is_compressed
from .compress_utils import extract_from_compressed

from .download_utils import decide_extension
from .download_utils import download_file
from .download_utils import src2base64
from .download_utils import path2base64

from .driver_utils import DRIVER_TYPE
from .driver_utils import get_chrome_options
from .driver_utils import get_driver
from .driver_utils import try_find_element
from .driver_utils import try_find_element_click
from .driver_utils import try_find_element_send_keys
from .driver_utils import pass_forms
from .driver_utils import click
from .driver_utils import download_PDF_with_driver
from .driver_utils import wait_until_all_elements
from .driver_utils import scrollDown

from .environ_utils import name2envname
from .environ_utils import where_is_envfile
from .environ_utils import read_environ
from .environ_utils import write_environ
from .environ_utils import show_environ
from .environ_utils import load_environ
from .environ_utils import check_environ

from .generic_utils import handleKeyError
from .generic_utils import handleTypeError
from .generic_utils import mk_class_get
from .generic_utils import recreate_dir
from .generic_utils import readable_bytes
from .generic_utils import now_str
from .generic_utils import splitted_query_generator
from .generic_utils import get_latest_filename
from .generic_utils import DictParamProcessor
from .generic_utils import str_strip
from .generic_utils import try_wrapper

from .journal_utils import whichJournal
from .journal_utils import canonicalize

from .monitor_utils import progress_reporthook_create
from .monitor_utils import ProgressMonitor

from .outfmt_utils import sanitize_filename
from .outfmt_utils import get_jinja_all_attrs
from .outfmt_utils import check_contents
from .outfmt_utils import tohtml
from .outfmt_utils import html2pdf
from .outfmt_utils import toPDF

from .pdf_utils import get_pdf_pages
from .pdf_utils import parser_pdf_pages
from .pdf_utils import get_pdf_contents

from .soup_utils import str2soup
from .soup_utils import split_section
from .soup_utils import group_soup_with_head
from .soup_utils import replace_soup_tag
from .soup_utils import find_target_text
from .soup_utils import find_target_id


#coding: utf-8
""" Utility programs for :mod:`journals <gummy.journals>` """
import os
import re
import sys
import time
import warnings
import requests
from bs4 import BeautifulSoup

from ._exceptions import JournalTypeIndistinguishableError
from .coloring_utils import toRED, toBLUE, toGREEN, toACCENT

DOMAIN2JOURNAL = {
    "academic.oup.com"                          : "OxfordAcademic",
    "advances.sciencemag.org"                   : "ScienceAdvances",
    "agupubs.onlinelibrary.wiley.com"           : "AGUPublications",
    "ajp.amjpathol.org"                         : "ASIP",
    "anatomypubs.onlinelibrary.wiley.com"       : "AnatomyPubs",
    "arxiv.org"                                 : "arXiv",
    "bio.biologists.org"                        : "Biologists",
    "biologydirect.biomedcentral.com"           : "BioMedCentral",
    "bmcbioinformatics.biomedcentral.com"       : "BioMedCentral",
    "bmcevolbiol.biomedcentral.com"             : "BioMedCentral",
    "bmcgenomics.biomedcentral.com"             : "BioMedCentral",
    "dev.biologists.org"                        : "Biologists",
    "dl.acm.org"                                : "ACM",
    "febs.onlinelibrary.wiley.com"              : "WileyOnlineLibrary",
    "genesdev.cshlp.org"                        : "GeneDev",
    "ieeexplore.ieee.org"                       : "ieeexplore",
    "iovs.arvojournals.org"                     : "ARVOJournals",
    "jamanetwork.com"                           : "JAMANetwork",
    "jcs.biologists.org"                        : "Biologists",
    "jkms.org"                                  : "JKMS",
    "journals.ametsoc.org"                      : "AMS",
    "journals.aps.org"                          : "APS",
    "journals.lww.com"                          : "LWWJournals",
    "journals.physiology.org"                   : "RenalPhysiology",
    "journals.plos.org"                         : "PLOSONE",
    "journals.sagepub.com"                      : "SAGEjournals",
    "jov.arvojournals.org"                      : "ARVOJournals",
    "keio.pure.elsevier.com"                    : "UniKeio",
    "learnmem.cshlp.org"                        : "LearningMemory",
    "link.springer.com"                         : "Springer",
    "linkinghub.elsevier.com"                   : "ScienceDirect",
    "mcb.asm.org"                               : "MolCellBio",
    "onlinelibrary.wiley.com"                   : "WileyOnlineLibrary",
    "pubmed.ncbi.nlm.nih.gov"                   : "PubMed",
    "pubs.acs.org"                              : "ACSPublications",
    "pubs.rsc.org"                              : "RSCPublishing",
    "pubs.rsna.org"                             : "RadioGraphics",
    "retrovirology.biomedcentral.com"           : "BioMedCentral",
    "rnajournal.cshlp.org"                      : "RNAjournal",
    "science.sciencemag.org"                    : "ScienceMag",
    "stemcellsjournals.onlinelibrary.wiley.com" : "StemCells",
    "tvst.arvojournals.org"                     : "ARVOJournals",
    "www.aclweb.org"                            : "ACLAnthology",
    "www.biorxiv.org"                           : "bioRxiv",
    "www.bioscience.org"                        : "Bioscience",
    "www.cell.com"                              : "CellPress",
    "www.e-ce.org"                              : "ClinicalEndoscopy",
    "www.embopress.org"                         : "EMBOPress",
    "www.genetics.org"                          : "Genetics",
    "www.frontiersin.org"                       : "frontiers",
    "www.intechopen.com"                        : "IntechOpen",
    "www.jbc.org"                               : "JBC",
    "www.jkms.org"                              : "JKMS",
    "www.jkns.or.kr"                            : "JKNS",
    "www.jpedsurg.org"                          : "PediatricSurgery",
    "www.jsse.org"                              : "JSSE",
    "www.jstage.jst.go.jp"                      : "JSTAGE",
    "www.lungcancerjournal.info"                : "LungCancer",
    "www.medrxiv.org"                           : "medRxiv",
    "www.mdpi.com"                              : "MDPI",
    "www.nature.com"                            : "Nature",
    "www.ncbi.nlm.nih.gov"                      : "NCBI",
    "www.nejm.org"                              : "NEJM",
    "www.nrcresearchpress.com"                  : "NRCResearchPress",
    "www.oncotarget.com"                        : "Oncotarget",
    "www.ou.edu"                                : "UniOKLAHOMA",
    "www.plantphysiol.org"                      : "ASPB",
    "www.pnas.org"                              : "PNAS",
    "www.psychiatrist.com"                      : "PsyChiArtist",
    "www.sciencedirect.com"                     : "ScienceDirect",
    "www.spandidos-publications.com"            : "Spandidos",
    "www.tandfonline.com"                       : "TaylorandFrancisOnline",
}

def canonicalize(url, driver=None, sleep_for_loading=1):
    """canonicalize the URL by accessing the URL once.

    Args:
        url (str)               : URL of the paper.
        driver (WebDriver)      : Selenium WebDriver. (default= ``None``)
        sleep_for_loading (int) : Number of seconds to wait for a web page to load (default= ``1`` )
    
    Returns:
        str : canonized URL.
    """

    if driver is not None:
        driver.get(url)
        time.sleep(sleep_for_loading)
        cano_url = driver.current_url
    else:
        try:
            ret = requests.get(url=url)
            cano_url = ret.url
        except:
            cano_url = url
    return cano_url

def whichJournal(url, driver=None, verbose=True):
    """ Decide which journal from the domain of the ``url``

    If the ``journal_type`` cannot be determined, Twitter DM |twitter badge| will 
    open automatically, so please feel free to request it to the developer.

    Args:
        url (str)          : URL of the paper.
        driver (WebDriver) : Selenium WebDriver. (default= ``None``)
        verbose (bool)     : Whether to print message or not. (default= ``True``) 

    Returns:
        str : journal_type

    Examples:
        >>> journal_type = whichJournal("https://www.ncbi.nlm.nih.gov/pmc/articles/PMC1573881/")
        Estimated Journal Type : NCBI
        >>> journal_type
        'ncbi'

    .. |twitter badge| image:: https://img.shields.io/badge/twitter-Requests-1da1f2?style=flat-square&logo=twitter
        :target: https://www.twitter.com/messages/compose?recipient_id=1042783905697288193&text=Please%20support%20this%20journal%3A%20
    """
    ext = os.path.splitext(url)[-1]
    url = canonicalize(url, driver=driver)
    if url.startswith("data:"):
        journal_type = "pdf"
    else:
        url_domain = re.match(pattern=r"^https?:\/\/(.+?)\/", string=url).group(1)
        journal_type = DOMAIN2JOURNAL.get(url_domain)
        if journal_type is None:
            if ext == ".pdf":
                journal_type = "pdf"
            else:
                msg = f"""
                {toGREEN('gummy.utils.journal_utils.whichJournal')} could not distinguish the journal type.
                * Please send a DM to the developer to support this journal ({toBLUE(url)})
                * Please specify the {toBLUE('journal_type')} explicitly until it is supported.
                * {toRED('I would really appreciate it if you could send a pull request.')}
                """
                raise JournalTypeIndistinguishableError(msg=msg, url=url)
    if verbose: print(f"Estimated Journal Type : {toACCENT(journal_type)}")
    return journal_type.lower()


# coding: utf-8

# Journal Data.
JournalData = {
	"Nature"                 : ["https://www.nature.com/articles/ncb0800_500",],
	"NCBI"                   : ["https://www.ncbi.nlm.nih.gov/pmc/articles/PMC1573881/", "https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3682417/"],
	"PubMed"                 : ["https://pubmed.ncbi.nlm.nih.gov/15125233/",],
	"OxfordAcademic"         : ["https://academic.oup.com/endo/article/144/11/5118/2881228",],
	"ScienceDirect"          : ["https://www.sciencedirect.com/science/article/abs/pii/S0006291X00924586?via%3Dihub",],
	"Springer"               : ["https://link.springer.com/article/10.1023/A:1009641613826",],
	"MDPI"                   : ["https://www.mdpi.com/1422-0067/14/6/11171",],
	"UniOKLAHOMA"            : ["https://www.ou.edu/journals/dis/DIS83/Technique83/7tecUiTei/UiTei.htm",],
	"LungCancer"             : ["https://www.lungcancerjournal.info/article/S0169-5002(02)00218-0/fulltext",],
	"CellPress"              : ["https://www.cell.com/fulltext/S0960-9822(02)01394-5",],
	"WileyOnlineLibrary"     : ["https://onlinelibrary.wiley.com/doi/abs/10.1002/arch.20006","https://febs.onlinelibrary.wiley.com/doi/full/10.1016/S0014-5793%2800%2901883-4",],
	"JBC"                    : ["https://www.jbc.org/content/280/13/12967",],
	"Biologists"             : ["https://dev.biologists.org/content/132/11/2547", "https://jcs.biologists.org/content/127/8/1805", "https://bio.biologists.org/content/9/2/bio050435",],
	"BioMedCentral"          : ["https://bmcbioinformatics.biomedcentral.com/articles/10.1186/1471-2105-10-392",],
	"ieeexplore"             : ["https://ieeexplore.ieee.org/document/4752474",],
	"JSTAGE"                 : ["https://www.jstage.jst.go.jp/article/jnms/77/2/77_2_71/_article",],
	"ACSPublications"        : ["https://pubs.acs.org/doi/10.1021/nl101829m",],
	"StemCells"              : ["https://stemcellsjournals.onlinelibrary.wiley.com/doi/full/10.1002/stem.615",],
	"UniKeio"                : ["https://keio.pure.elsevier.com/ja/publications/enhanced-specificity-of-hpv16-e6e7-sirna-by-rna-dna-chimera-modif",],
	"PLOSONE"                : ["https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0028688",],
	"frontiers"              : ["https://www.frontiersin.org/articles/10.3389/fgene.2012.00101/full",],
	"RNAjournal"             : ["https://rnajournal.cshlp.org/content/19/1/17",],
	"IntechOpen"             : ["https://www.intechopen.com/books/gene-expression-and-regulation-in-mammalian-cells-transcription-from-general-aspects/current-status-for-application-of-rna-interference-technology-as-nucleic-acid-drug",],
	"NRCResearchPress"       : ["https://www.nrcresearchpress.com/doi/10.1139/gen-2016-0127",],
	"Spandidos"              : ["https://www.spandidos-publications.com/10.3892/ijmm.2015.2141",],
	"TaylorandFrancisOnline" : ["https://www.tandfonline.com/doi/full/10.1080/15476286.2018.1486658",],
	"bioRxiv"                : ["https://www.biorxiv.org/content/10.1101/724575v1",],
	"RSCPublishing"          : ["https://pubs.rsc.org/en/content/articlelanding/2004/CS/b303644h",],
	"JSSE"                   : ["https://www.jsse.org/index.php/jsse/article/view/531/528",],
	"ScienceAdvances"        : ["https://advances.sciencemag.org/content/6/20/eaaz6485",],
	"medRxiv"                : ["https://www.medrxiv.org/content/10.1101/2020.04.21.20074567v2",],
	"ACLAnthology"           : ["https://www.aclweb.org/anthology/2020.acl-main.28",],
	"PNAS"                   : ["https://www.pnas.org/content/113/20/5498",],
	"AMS"                    : ["https://journals.ametsoc.org/mwr/article/146/7/1965/103117",],
	# "ACM"                    : ["https://dl.acm.org/doi/abs/10.1145/2470654.2466191",],
	"APS"                    : ["https://journals.aps.org/prresearch/abstract/10.1103/PhysRevResearch.2.033237",],
	"ASIP"                   : ["https://ajp.amjpathol.org/article/S0002-9440(10)65401-6/fulltext",],
	"AnatomyPubs"            : ["https://anatomypubs.onlinelibrary.wiley.com/doi/full/10.1002/dvdy.22501",],
	"RenalPhysiology"        : ["https://journals.physiology.org/doi/full/10.1152/ajprenal.00262.2009",],
	"Genetics"               : ["https://www.genetics.org/content/176/4/2177",],
	"GeneDev"                : ["http://genesdev.cshlp.org/content/5/4/697",],
	"JAMANetwork"            : ["https://jamanetwork.com/journals/jamaotolaryngology/fullarticle/1921366",],
	"SAGEjournals"           : ["https://journals.sagepub.com/doi/10.1597/1545-1569_1995_032_0251_saips_2.3.co_2",],
	"MolCellBio"             : ["https://mcb.asm.org/content/25/2/671",],
	"JKMS"                   : ["https://jkms.org/DOIx.php?id=10.3346/jkms.2006.21.2.374",],
	"JKNS"                   : ["https://www.jkns.or.kr/journal/view.php?doi=10.3340/jkns.2016.59.3.187",],
	"Bioscience"             : ["https://www.bioscience.org/2009/v1s/af/12/fulltext.htm",],
	"RadioGraphics"          : ["https://pubs.rsna.org/doi/10.1148/rg.2017170017",],
	"PediatricSurgery"       : ["https://www.jpedsurg.org/article/S0022-3468(99)90787-X/pdf",],
	"AGUPublications"        : ["https://agupubs.onlinelibrary.wiley.com/doi/full/10.1029/2019GL084683",],
	"NEJM"                   : ["https://www.nejm.org/doi/full/10.1056/NEJMcpc1913472",],
	"LWWJournals"            : ["https://journals.lww.com/ajsp/Fulltext/2020/09000/Thyroid_Papillary_Microtumor__Validation_of_the.2.aspx",],
	"ARVOJournals"           : ["https://iovs.arvojournals.org/article.aspx?articleid=2770586#247080114", "https://jov.arvojournals.org/Article.aspx?articleid=2766199", "https://tvst.arvojournals.org/Article.aspx?articleid=2766271",],
	"LearningMemory"         : ["http://learnmem.cshlp.org/content/19/9/359.full.html",],
	"ScienceMag"             : ["https://science.sciencemag.org/content/359/6371/108",],
	"PsyChiArtist"           : ["https://www.psychiatrist.com/JCP/article/Pages/cbt-for-pharmacotherapy-resistant-depression.aspx"],
	"Oncotarget"             : ["https://doi.org/10.18632/oncotarget.21129",],
	"ClinicalEndoscopy"      : ["https://www.e-ce.org/journal/view.php?doi=10.5946/ce.2020.081"],
	"ASPB"                   : ["http://www.plantphysiol.org/content/125/4/1688",],
	"EMBOPress"              : ["https://www.embopress.org/doi/abs/10.1002/j.1460-2075.1996.tb01033.x",],
}
JournalData = {journal.lower():url for journal,url in JournalData.items()}

# Sentence Data
SentenceData = dict(
    simple_sentence    = "This is a pen.",
    multiple_sentences = "I have a pen. I have an apple. Ah, apple pen.",
    SteveJobsSpeech    = "Thank You. I am honored to be with you today for your commencement from one of the finest universities in the world. Truth be told I never graduated from college and this is the closest I’ve ever gotten to a college graduation. Today I want to tell you three stories from my life. That’s it. No big deal. Just three stories. The first story is about connecting the dots. I dropped out of Reed College after the first 6 months, but then stayed around as a drop-in for another 18 months or so before I really quit. So why did I drop out? It started before I was born. My biological mother was a young, unwed graduate student, and she decided to put me up for adoption. She felt very strongly that I should be adopted by college graduates, so everything was all set for me to be adopted at birth by a lawyer and his wife. Except that when I popped out they decided at the last minute that they really wanted a girl. So my parents, who were on a waiting list, got a call in the middle of the night asking: “We’ve got an unexpected baby boy; do you want him?” They said: “Of course.” My biological mother found out later that my mother had never graduated from college and that my father had never graduated from high school. She refused to sign the final adoption papers. She only relented a few months later when my parents promised that I would go to college. This was the start in my life. And 17 years later I did go to college. But I naively chose a college that was almost as expensive as Stanford, and all of my working-class parents’ savings were being spent on my college tuition. After six months, I couldn’t see the value in it. I had no idea what I wanted to do with my life and no idea how college was going to help me figure it out. And here I was spending all of the money my parents had saved their entire life. So I decided to drop out and trust that it would all work out OK. It was pretty scary at the time, but looking back it was one of the best decisions I ever made. The minute I dropped out I could stop taking the required classes that didn’t interest me, and begin dropping in on the ones that looked far more interesting. It wasn’t all romantic. I didn’t have a dorm room, so I slept on the floor in friends’ rooms, I returned coke bottles for the 5 cent deposits to buy food with, and I would walk the 7 miles across town every Sunday night to get one good meal a week at the Hare Krishna temple. I loved it. And much of what I stumbled into by following my curiosity and intuition turned out to be priceless later on. Let me give you one example: Reed College at that time offered perhaps the best calligraphy instruction in the country. Throughout the campus every poster, every label on every drawer, was beautifully hand calligraphed. Because I had dropped out and didn’t have to take the normal classes, I decided to take a calligraphy class to learn how to do this. I learned about serif and sans-serif typefaces, about varying the amount of space between different letter combinations, about what makes great typography great. It was beautiful, historical, artistically subtle in a way that science can’t capture, and I found it fascinating. None of this had even a hope of any practical application in my life. But ten years later, when we were designing the first Macintosh computer, it all came back to me. And we designed it all into the Mac. It was the first computer with beautiful typography. If I had never dropped in on that single course in college, the Mac would have never had multiple typefaces or proportionally spaced fonts. And since Windows just copied the Mac, it’s likely that no personal computer would have them. If I had never dropped out, I would have never dropped in on that calligraphy class, and personal computers might not have the wonderful typography that they do. Of course it was impossible to connect the dots looking forward when I was in college. But it was very, very clear looking backward ten years later. Again, you can’t connect the dots looking forward; you can only connect them looking backward. So you have to trust that the dots will somehow connect in your future. You have to trust in something, your gut, destiny, life, karma, whatever. Because believing that the dots will connect down the road will give you the confidence to follow your heart. Even when it leads you off the well-worn path, and that will make all the difference."
)


# Save as db.
class TestData():
    def __init__(self):
        self.journals = JournalData
        self.sentences = SentenceData

    def insert_journal(self, journal, url):
        self.journals[journal] = url

    def get_journal(self, journal):
        return self.journals.get(journal)

    def iter_data(self, data_type):
        for key,value in self.__dict__.get(data_type).items():
            yield key,value



# coding: utf-8
"""These classes translate English into Japanese using an external 
translation service (website).

Currently, Translation-Gummy supports the following services:
    - `Google Translate <https://translate.google.co.jp/#en/ja/Translation%20Gummy>`_
    - `DeepL Translator <https://www.deepl.com/en/translator#en/ja/Translation%20Gummy>`_

You can easily get (import) ``Translator Class`` by the following ways.

.. code-block:: python

    >>> from gummy import translators
    >>> translator = translators.get("google")
    >>> translator
    <gummy.translators.GoogleTranslator at 0x129890160>
    >>> from gummy.translators import GoogleTranslator
    >>> google = GoogleTranslator()
    >>> google
    <gummy.translators.GoogleTranslator at 0x129890700>
    >>> translator = translators.get(google)
    >>> id(google) == id(translator)
    True
"""
import re
import time
import json
import urllib
import warnings
from abc import ABCMeta, abstractmethod, abstractproperty, abstractstaticmethod
from bs4 import BeautifulSoup

from .utils.coloring_utils import toBLUE, toGREEN
from .utils.driver_utils import get_driver
from .utils.generic_utils import handleKeyError, handleTypeError, mk_class_get, splitted_query_generator
from .utils.monitor_utils import ProgressMonitor
from .utils.soup_utils import find_target_text

class GummyAbstTranslator(metaclass=ABCMeta):
    def __init__(self, driver=None, maxsize=5000, interval=1, trials=30, verbose=False, use_cache=True):
        """If you want to create your own translator class, please inherit this class.

        Args:
            driver (WebDriver) : Selenium WebDriver.
            maxsize (int)      : Number of English characters that we can send a request at one time. (default= ``5000``)
            interval (int)     : Trial interval [s]. (default= ``1``)
            trials (int)       : How many times to try to find japanese text. (default= ``30``)
            verbose (bool)     : Whether to print message or not. (default= ``False``)
            use_cache (bool)   : Whether to use cache or not. cashe is used in :meth:`is_ja_enough <gummy.translators.GummyAbstTranslator.is_ja_enough>` (default= ``True``)

        Attributes:
            cache_ja (str) : Japanese acquired one time ago. Prevent bugs where the same Japanese is repeated. Used in :meth:`is_ja_enough <gummy.translators.GummyAbstTranslator.is_ja_enough>`.
        """
        self.driver = driver
        self.maxsize = maxsize
        self.interval = interval
        self.trials = trials
        self.verbose = verbose
        self.use_cache = use_cache
        self.cache_ja = ""

    @property
    def class_name(self):
        """Same as ``self.__class__.__name__``."""
        return self.__class__.__name__

    @property
    def name(self):
        """Translator service name."""
        return self.class_name.replace("Translator", "")

    @abstractproperty
    def en2ja_url_fmt(self):
        """Format of the query. English will be assigned to ``{english}``."""
        return "https://domain/english2japanese/?query={english}"

    @abstractstaticmethod
    def find_ja(soup):
        """Find translated Japanese text from ``soup``

        Args:
            soup (bs4.BeautifulSoup): A data structure representing a parsed HTML or XML document.

        Return:
            str: Japanese text.
        """
        return find_target_text(soup=soup, name="japanese")

    def is_ja_enough(self, ja):
        """Check if the acquired Japanese is appropriate

        Args:
            ja (str) : Translated Japanese text.
        
        Examples:
            >>> from gummy import translators
            >>> translator = translators.get("google")
            >>> translator.is_ja_enough("")
            False
            >>> translator.is_ja_enough("日本語")
            True
            >>> translator.cache_ja = "日本語"
            >>> translator.is_ja_enough("日本語")
            False
        """
        return (len(ja)>0) and (not self.cache_ja.startswith(ja))

    @property
    def driver_info(self):
        """Return driver_info

        Examples:
            >>> from gummy.utils import get_driver
            >>> from gummy.translators import GoogleTranslator
            >>> with get_driver() as driver:
            ...     translator = GoogleTranslator(driver=driver)
            ...     print(translator.driver_info)
            {'session_id': '3e869e62f06c5f7d938831179ad3c50e', 'browserName': 'chrome'}
            >>> translator = GoogleTranslator(driver=None)
            >>> print(translator.driver_info)
            {}
        """
        info = {}
        driver = self.driver
        if driver is not None:
            info["session_id"] = driver.session_id
            info["browserName"] = driver.capabilities.get("browserName")
        return info

    def check_en2ja(self):
        """Make sure :meth:`en2ja_url_fmt <gummy.translators.GummyAbstTranslator.en2ja_url_fmt>` is appropriate.

        Raises:
            TypeError  : When ``self.en2ja_url_fmt`` is not ``str`` type.
            ValueError : When ``{english}`` not in ``self.en2ja_url_fmt``.
        """
        # en2ja format.
        if not isinstance(self.en2ja_url_fmt, str):
            raise TypeError(f"`self.en2ja_url_fmt` must be str not {type(self.en2ja_url_fmt)}")
        if self.en2ja_url_fmt.find("{english}") == -1:
            raise ValueError("Please include {english} in `self.en2ja_url_fmt`")

    def check_driver(self, driver=None):
        """If the driver does not exist, use :meth:`get_driver <gummy.utils.driver_utils.get_driver>` to get the driver."""
        driver = driver or self.driver
        if driver is None:
            driver = get_driver()
        self.driver = driver
        if self.verbose: print(f"Driver info:\n{json.dumps(self.driver_info, indent=2)}")
        return driver

    def en2ja(self, query, driver=None, barname=None):
        """Translate English into Japanese.

        Args:
            query (str)        : English to be translated.
            driver (WebDriver) : Selenium WebDriver.
            barname (str)      : Bar name for :meth:`ProgressMonitor <gummy.utils.monitor_utils.ProgressMonitor>`.

        Examples:
            >>> from gummy.translators import GoogleTranslator
            >>> ja = translator.en2ja("This is a pen.")
            GoogleTranslator (query1) 01/30 [--------------------]  3.33% - 2.731[s]
            >>> print(ja)
            これはペンです。
            >>> ja = translator.en2ja("This is a pen.", barname="^_^")
            ^_^ (query1) 30/30 [####################]100.00% - 81.073[s]
            >>> print(ja)
            これはペンです。
            >>> # The reason why it takes a long time above is that the same Japanese is repeated.
            >>> translator.cache_ja = ""
            >>> ja = translator.en2ja("This is a pen.", barname="^_^")
            ^_^ (query1) 01/30 [--------------------]  3.33% - 2.725[s]
            >>> print(ja)
            これはペンです。
        """
        self.check_en2ja()
        driver = self.check_driver(driver=driver)
        maxsize = self.maxsize
        interval = self.interval
        trials = self.trials
        verbose = self.verbose
        if barname is None: barname = self.class_name
        
        japanese = []
        gen = splitted_query_generator(query=query, maxsize=maxsize)
        for i,q in enumerate(gen):
            url = self.en2ja_url_fmt.format(english=urllib.parse.quote(q))
            driver.refresh()
            driver.get(url)
            monitor = ProgressMonitor(max_iter=trials, verbose=verbose, barname=f"{barname} (query{i+1})")
            for i in range(trials):
                time.sleep(interval)
                html = driver.page_source.encode("utf-8")
                soup = BeautifulSoup(html, "lxml")
                ja = self.find_ja(soup)
                monitor.report(i, japanese=ja)
                if self.is_ja_enough(ja):
                    break
            monitor.remove()
            japanese.append(ja)
            if self.use_cache:
                self.cache_ja = ja
            time.sleep(1)
        
        japanese = "".join(japanese)
        return japanese

class DeepLTranslator(GummyAbstTranslator):
    """
    DeepL is a deep learning company that develops artificial intelligence systems 
    for languages. See https://www.deepl.com/en/home for more info.
    """
    def __init__(self, driver=None, maxsize=5000, interval=1, trials=30, verbose=False, use_cache=True):
        super().__init__(driver=driver, maxsize=maxsize, interval=interval, trials=trials, verbose=verbose, use_cache=use_cache)

    @property
    def en2ja_url_fmt(self):
        return "https://www.deepl.com/en/translator#en/ja/{english}"

    @staticmethod
    def find_ja(soup):
        return find_target_text(soup=soup, name="button", class_="lmt__translations_as_text__text_btn")

    def is_ja_enough(self, ja):
        """Deepl represents the character being processed as ``[...]``, so make sure it isn't at the end of the japanese.
        
        Examples:
            >>> from gummy import translators
            >>> translator = translators.get("deepl")
            >>> translator.is_ja_enough("日本語 [...]")
            False
            >>> translator.is_ja_enough("[...]日本語")
            True
            >>> translator.is_ja_enough("日本語[...]日本語")
            True
        """
        return super().is_ja_enough(ja) and (not ja.endswith("[...]"))

class GoogleTranslator(GummyAbstTranslator):
    """Google Translate is a free multilingual neural machine translation service 
    developed by Google, to translate text and websites from one language into 
    another. See https://translate.google.com/ for more info.
    """
    def __init__(self, driver=None, maxsize=5000, interval=1, trials=30, verbose=False, use_cache=True):
        super().__init__(driver=driver, maxsize=maxsize, interval=interval, trials=trials, verbose=verbose, use_cache=use_cache)

    @property
    def en2ja_url_fmt(self):
        return "https://translate.google.co.jp/#en/ja/{english}"

    @staticmethod
    def find_ja(soup):
        return find_target_text(soup=soup, name="span", class_="tlid-translation translation", attrs={"lang": "ja"})

all = TranslationGummyTranslators = {
    "google" : GoogleTranslator,
    "deepl"  : DeepLTranslator,
}

get = mk_class_get(
    all_classes=TranslationGummyTranslators,
    gummy_abst_class=[GummyAbstTranslator],
    genre="translators",
)


#coding: utf-8

def test_sentences(db):
    simple_sentence    = db.sentences.get("simple_sentence")
    multiple_sentences = db.sentences.get("multiple_sentences")
    SteveJobsSpeech    = db.sentences.get("SteveJobsSpeech")

    assert simple_sentence    is not None
    assert multiple_sentences is not None
    assert SteveJobsSpeech    is not None

    assert len(simple_sentence)    < len(multiple_sentences)
    assert len(multiple_sentences) < len(SteveJobsSpeech)


# coding: utf-8
""" Utility programs for monitoring loop or time-consuming process. """
import sys
import time

from .coloring_utils import toACCENT, toBLUE
from .generic_utils import readable_bytes

def progress_reporthook_create(filename="", bar_width=20, verbose=True):
    """Create Progress reporthook for ``urllib.request.urlretrieve``

    Returns:
        The ``reporthook`` which is a callable that accepts a ``block number``, a ``read size``, and the ``total file size`` of the URL target.

    Args:
        filename (str)  : Downloading filename.
        bar_width (int) : The width of progress bar.

    Examples:
        >>> import urllib
        >>> from gummy.utils import progress_reporthook_create
        >>> urllib.request.urlretrieve(url="hoge.zip", filename="hoge.zip", reporthook=progress_reporthook_create(filename="hoge.zip"))
        hoge.zip	1.5%[--------------------] 21.5[s] 8.0[GB/s]	eta 1415.1[s]
    """
    def progress_reporthook_verbose(block_count, block_size, total_size):
        global _reporthook_start_time
        if block_count == 0:
            _reporthook_start_time = time.time()
            return
        progress_size = block_count*block_size
        percentage = min(1.0, progress_size/total_size)
        progress_bar = ("#" * int(percentage * bar_width)).ljust(bar_width, "-")
        
        duration = time.time() - _reporthook_start_time
        speed = progress_size / duration
        eta = (total_size-progress_size)/speed

        speed, speed_unit = readable_bytes(speed)
        
        sys.stdout.write(f"\r{filename}\t{percentage:.1%}[{progress_bar}] {duration:.1f}[s] {speed:.1f}[{speed_unit}/s]\teta {eta:.1f}[s]")
        if progress_size >= total_size: print()
    def progress_reporthook_non_verbose(block_count, block_size, total_size):
        pass
    return progress_reporthook_verbose if verbose else progress_reporthook_non_verbose

class ProgressMonitor():
    """Monitor the loop progress.

    Examples:
        >>> from pycharmers.utils import ProgressMonitor
        >>> max_iter = 100
        >>> monitor = ProgressMonitor(max_iter=max_iter, verbose=1, barname="NAME")
        >>> for it in range(max_iter):
        >>>     monitor.report(it, loop=it+1)
        >>> monitor.remove()
        NAME 100/100[####################]100.00% - 0.010[s]  loop: 100
    """
    def __init__(self, max_iter, verbose=1, barname="", **kwargs):
        """
        Args:
            max_iter (int) : Maximum number of iterations.
            verbose (int)  : -1, 0, 1
                - -1 : silent
                - 0  : only progress bar
                - 1  : progress bar and metrics
                - 2  : progress plot
            barname (str)  : barname
        """
        self._init()
        self.max_iter = max_iter
        self.digit = len(str(max_iter))
        self.verbose = verbose
        self.barname = barname + " " if len(barname)>0 else ""
        self.report = {
            -1 : self._report_silent,
             0 : self._report_only_prograss_bar,
             1 : self._report_progress_bar_and_metrics,
        }.get(verbose, self._report_progress_bar_and_metrics)
        self.report(it=-1)

    def _init(self):
        self.histories = {}
        self.iter = 0
        self.initial_seconds_since_epoch = time.time()

    def _report_silent(self, it, **metrics):
        pass

    def _report_only_prograss_bar(self, it, **metrics):
        it += 1
        sys.stdout.write(
            f"\r{self.barname}{it:>0{self.digit}}/{self.max_iter} " + \
            f"[{('#' * int((it/self.max_iter)/0.05)).ljust(20, '-')}]" + \
            f"{it/self.max_iter:>7.2%} - {time.time()-self.initial_seconds_since_epoch:.3f}[s]"
        )

    def _report_progress_bar_and_metrics(self, it, **metrics):
        it += 1
        metric = ", ".join([f"{toACCENT(k)}: {toBLUE(v)}" for  k,v in metrics.items()])
        sys.stdout.write(
            f"\r{self.barname}{it:>0{self.digit}}/{self.max_iter}" + \
            f"[{('#' * int((it/self.max_iter)/0.05)).ljust(20, '-')}]" + \
            f"{it/self.max_iter:>7.2%} - {time.time()-self.initial_seconds_since_epoch:.3f}[s]   " + \
            f"{metric}"
        )

    def remove(self):
        """Do the necessary processing at the end."""
        def _pass():
            pass
        {
            -1: _pass,
             0: print,
             1: print,
        }.get(self.verbose, print)()


# coding: utf-8
"""This file defines a model that integrates all of :mod:`journals <gummy.journals>`, 
:mod:`translators <gummy.translators>`, :mod:`gateways <gummy.gateways>`, and 
it is possible to do all of the following at once.

1. Determine the ``journal_type`` of paper from the ``url`` or file extension.
2. If necessary, use a ``GummyGateway`` to access non-open content of the journal.
3. Parse the paper using ``GummyJournals`` and obtain the contents.
4. Translate the English obtained using ``GummyTranslators`` to Japanese.
5. Arrange Japanese and English according to the `templates <https://github.com/iwasakishuto/Translation-Gummy/tree/master/gummy/templates>`_ .
6. Convert the obtained HTML to PDF.

You can get (import) ``TranslationGummy`` by the following 2 ways.

.. code-block:: python

    >>> from gummy.models import TranslationGummy
    >>> from gummy import TranslationGummy
"""

import os
import time

from . import gateways
from . import journals
from . import translators

from .utils._path import GUMMY_DIR, TEMPLATES_DIR
from .utils.coloring_utils import toACCENT, toBLUE, toGREEN
from .utils.driver_utils import get_driver
from .utils.journal_utils import whichJournal
from .utils.outfmt_utils import tohtml, html2pdf, sanitize_filename
from .utils.driver_utils import get_driver

class TranslationGummy():
    """This class integrates all of the followings
    
    - :mod:`journals <gummy.journals>`
    - :mod:`translators <gummy.translators>`
    - :mod:`gateways <gummy.gateways>`

    Args:
        chrome_options (ChromeOptions)    : Instance of ChromeOptions. (default= :meth:`get_chrome_options() <gummy.utils.driver_utils.get_chrome_options>` ) 
        browser (bool)                    : Whether you want to run Chrome with GUI browser. (default= ``False`` ) 
        driver (WebDriver)                : Selenium WebDriver.
        gateway (str, GummyGateWay)       : identifier of the Gummy Gateway Class. See :mod:`gateways <gummy.gateways>`. (default= `"useless"`)
        translator (str, GummyTranslator) : identifier of the Gummy Translator Class. See :mod:`translators <gummy.translators>`. (default= `"deepl"`)
        maxsize (int)                     : Number of English characters that we can send a request at one time. (default= ``5000``)
        verbose (bool)                    : Whether you want to print output or not. (default= ``True`` ) 
        translator_verbose (bool)         : Whether you want to print translator’s output or not. (default= ``False`` ) 
    """
    def __init__(self, chrome_options=None, browser=False, driver=None, 
                 gateway="useless", translator="deepl", maxsize=5000, 
                 verbose=True, translator_verbose=False):
        if driver is None: driver = get_driver(chrome_options=chrome_options, browser=browser)
        self.driver = driver
        self.gateway = gateway
        self.translator = translators.get(translator, maxsize=maxsize, verbose=translator_verbose)
        self.verbose = verbose

    def en2ja(self, query, barname=None):
        """Translate English into Japanese. See :meth:`en2ja <gummy.translators.en2ja>`.

        Args:
            query (str)        : English to be translated.
            barname (str)      : Bar name for :meth:`ProgressMonitor <gummy.utils.monitor_utils.ProgressMonitor>`.

        Examples:
            >>> from gummy import TranslationGummy
            >>> model = TranslationGummy()
            >>> ja = model.en2ja("This is a pen.")
            DeepLTranslator (query1) 03/30 [##------------------] 10.00% - 3.243[s]
            >>> ja
            'これはペンです。'
        """
        return self.translator.en2ja(query=query, driver=self.driver, barname=barname)

    def get_contents(self, url, journal_type=None, crawl_type=None, gateway=None, **gatewaykwargs):
        """Get contents of the journal.

        Args:
            url (str)                   : URL of a paper or ``path/to/local.pdf``.
            journal_type (str)          : Journal type, if you not specify, judge by analyzing from ``url``.
            crawl_type (str)            : Crawling type, if you not specify, use recommended crawling type.
            gateway (str, GummyGateWay) : identifier of the Gummy Gateway Class. See :mod:`gateways <gummy.gateways>`. (default= ``None``)
            gatewaykwargs (dict)        : Gateway keywargs. See :meth:`passthrough <gummy.gateways.GummyAbstGateWay.passthrough>`.

        Returns:
            tuple (str, dict) : (title, content)

        Examples:
            >>> from gummy import TranslationGummy
            >>> model = TranslationGummy()
            >>> title, texts = model.get_contents("https://www.nature.com/articles/ncb0800_500")
            Estimated Journal Type : Nature
            Crawling Type: soup
                :
            >>> print(title)
            Formation of the male-specific muscle in female by ectopic expression
            >>> print(texts[:1])
            [{'head': 'Abstract', 'en': 'The  () gene product Fru has been ... for the sexually dimorphic actions of the gene.'}]
        """
        if journal_type is None:
            if os.path.exists(url):
                journal_type = "pdf"
            else:
                journal_type = whichJournal(url, driver=self.driver, verbose=self.verbose)
        gateway = gateway or self.gateway
        crawler = journals.get(journal_type, gateway=gateway, sleep_for_loading=3, verbose=self.verbose)
        title, texts = crawler.get_contents(url=url, driver=self.driver, crawl_type=crawl_type, **gatewaykwargs)
        return title, texts

    def toHTML(self, url, path=None, out_dir=GUMMY_DIR,
               journal_type=None, crawl_type=None, gateway=None,
               searchpath=TEMPLATES_DIR, template="paper.html", 
               **gatewaykwargs):
        """Get contents from URL and create a HTML.

        Args:
            url (str)                   : URL of a paper or ``path/to/local.pdf``.
            path/out_dir (str)          : Where you save a created HTML. If path is None, save at ``<out_dir>/<title>.html`` (default= `GUMMY_DIR`)
            journal_type (str)          : Journal type, if you specify, use ``journal_type`` journal crawler. (default= `None`)
            crawl_type (str)            : Crawling type, if you not specify, use recommended crawling type. (default= `None`)
            gateway (str, GummyGateWay) : identifier of the Gummy Gateway Class. See :mod:`gateways <gummy.gateways>`. (default= `None`)
            searchpath/template (str)   : Use a ``<searchpath>/<template>`` tpl for creating HTML. (default= `TEMPLATES_DIR/paper.html`)
            gatewaykwargs (dict)        : Gateway keywargs. See :meth:`passthrough <gummy.gateways.GummyAbstGateWay.passthrough>`.
        """
        title, contents = self.get_contents(
            url=url, journal_type=journal_type, crawl_type=crawl_type, 
            gateway=gateway, **gatewaykwargs
        )
        print(f"\nTranslation: {toACCENT(self.translator.name)}\n{'='*30}")
        len_contents = len(contents)
        for i,content in enumerate(contents):
            barname = f"[{i+1:>0{len(str(len_contents))}}/{len_contents}] " + toACCENT(content.get("head","\t"))            
            if "en" in content:
                # ===== TRANSLATION ======
                ja = self.en2ja(query=content["en"], barname=barname)
                content["ja"] = ja
                # ========================
            elif "img" in content and self.verbose:
                print(barname + "<img>")
        if path is None:
            path = os.path.join(out_dir, sanitize_filename(fp=title, dirname="."))
        htmlpath = tohtml(
            path=path, title=title, contents=contents, 
            searchpath=searchpath, template=template, verbose=self.verbose
        )
        return htmlpath

    def toPDF(self, url, path=None, out_dir=GUMMY_DIR,
              journal_type=None, crawl_type=None, gateway=None, 
              searchpath=TEMPLATES_DIR, template="paper.html",
              delete_html=True, options={}, 
              **gatewaykwargs):
        """Get contents from URL and create a PDF.

        Args:
            url (str)                   : URL of a paper or ``path/to/local.pdf``.
            path/out_dir (str)          : Where you save a created HTML. If path is None, save at ``<out_dir>/<title>.html`` (default= `GUMMY_DIR`)
            journal_type (str)          : Journal type, if you specify, use ``journal_type`` journal crawler. (default= `None`)
            crawl_type (str)            : Crawling type, if you not specify, use recommended crawling type. (default= `None`)
            gateway (str, GummyGateWay) : identifier of the Gummy Gateway Class. See :mod:`gateways <gummy.gateways>`. (default= `None`)
            searchpath/template (str)   : Use a ``<searchpath>/<template>`` tpl for creating HTML. (default= `TEMPLATES_DIR/paper.html`)
            delete_html (bool)          : Whether you want to delete an intermediate html file. (default= `True`) 
            options (dict)              : Options for wkhtmltopdf. See https://wkhtmltopdf.org/usage/wkhtmltopdf.txt (default= `{}`)
            gatewaykwargs (dict)        : Gateway keywargs. See :meth:`passthrough <gummy.gateways.GummyAbstGateWay.passthrough>`.
        """
        htmlpath = self.toHTML(
            url=url, path=path, out_dir=out_dir,
            journal_type=journal_type, crawl_type=crawl_type, gateway=gateway, 
            searchpath=searchpath, template=template,
            **gatewaykwargs
        )
        if self.verbose: print(f"\nConvert from HTML to PDF\n{'='*30}")
        pdfpath = html2pdf(path=htmlpath, delete_html=delete_html, verbose=self.verbose, options=options)
        return pdfpath


# coding: utf-8
"""CLI(Command Line Interface) tools

    Since the following two programs are frequently used, I created this file to be called from command line.

    - Translate your journals and generate a PDF.
    - Translation
"""
import sys
import argparse

from .models import TranslationGummy
from .journals import SUPPORTED_CRAWL_TYPES
from .utils._path import TEMPLATES_DIR, GUMMY_DIR
from .utils.driver_utils import get_chrome_options
from .utils.generic_utils import DictParamProcessor

def translate_journal(argv=sys.argv[1:]):
    """Translate journals.

    Args:
        url (str)                   : URL of a paper or ``path/to/local.pdf``. (required)
        -G/--gateway (str)          : Gateway identifier, string name of a gateway. (default= ``"useless"`` )
        -T/--translator (str)       : Translator identifier, string name of a translator. (default= ``"deepl"`` )
        -J/--journal (str)          : Journal identifier, string name of a journal. (default= ``None`` )
        --crawl-type (str)          : Crawling type, if you not specify, use recommended crawling type. (default= ``None`` )
        -O/--out-dir (str)          : Where you want to save a created PDF. (default= ``GUMMY_DIR`` )
        --browser (bool)            : Whether you want to run Chrome with GUI browser. (default= ``False`` )
        -pdf/--pdf-path (str)       : Path to output pdf file path. (default= ``None`` )
        -tpl/--tpl-path (str)       : Path to template path. (default= ``None`` )
        --save-html (bool)          : Whether you want to save an intermediate html file. (default= ``False`` )
        --quiet (bool)              : Whether you want to be quiet or not. (default= ``False`` )
        --translator-verbose (bool) : Whether you want to print translator's output or not. (default= ``False`` )
        -GP/--gateway-params (dict) : Specify the value required to pass through the gateway. You can specify by ``-GP username=USERNAME -GP password=PASSWORD`` (default= ``{}`` )

    Note:
        When you run from the command line, execute as follows::
        
        $ gummy-journal "https://www.nature.com/articles/ncb0800_500"

    Examples:
        >>> $ gummy-journal "https://www.nature.com/articles/ncb0800_500"
    """
    parser = argparse.ArgumentParser(prog="gummy-journal", add_help=True)
    parser.add_argument("url", type=str, help="URL of a page you want to create a pdf.")
    parser.add_argument("-G", "--gateway",    type=str, default="useless", help="Gateway identifier, string name of a gateway")
    parser.add_argument("-T", "--translator", type=str, default="deepl",   help="Translator identifier, string name of a translator")
    parser.add_argument("-J", "--journal",    type=str, default=None,      help="Journal identifier, string name of a journal")
    parser.add_argument("--crawl-type",       type=str, default=None,      help="Crawling type, if you not specify, use recommended crawling type.", choices=SUPPORTED_CRAWL_TYPES)
    parser.add_argument("-O", "--out-dir",    type=str, default=GUMMY_DIR, help="Where you want to save a created PDF.")
    # Chrome options
    parser.add_argument("--browser", action="store_true", help="Whether you want to run Chrome with GUI browser.")
    # PDF format
    parser.add_argument("-pdf", "--pdf-path", type=str, default=None, help="Path to output pdf file path.")
    parser.add_argument("-tpl", "--tpl-path", type=str, default=None, help="Path to template path.")
    parser.add_argument("--save-html",          action="store_true",  help="Whether you want to save an intermediate html file. (default=False)")
    parser.add_argument("--quiet",              action="store_true",  help="Whether you want to be quiet or not. (default=False)")
    parser.add_argument("--translator-verbose", action="store_true",  help="Whether you want to print translator's output or not. (default=False)")
    # Gateway kwargs
    parser.add_argument("-GP", "--gateway-params", default={}, action=DictParamProcessor, help="Specify the value required to pass through the gateway. You can specify by -GP username=USERNAME -GP password=PASSWORD")
    args = parser.parse_args(argv)

    chrome_options = get_chrome_options(browser=args.browser)
    url = args.url
    gateway = args.gateway
    translator = args.translator
    journal_type = args.journal
    crawl_type = args.crawl_type
    out_dir = args.out_dir

    pdf_path = args.pdf_path
    tpl_path = args.tpl_path
    delete_html = not args.save_html
    verbose = not args.quiet
    translator_verbose = args.translator_verbose
    gateway_params = args.gateway_params
    if tpl_path is None:
        searchpath = TEMPLATES_DIR
        template = "paper.html"
    else:
        *searchpath, template = tpl_path.split("/")
        searchpath = "/".join(searchpath) or "."

    model = TranslationGummy(
        chrome_options=chrome_options, gateway=gateway, translator=translator, 
        verbose=verbose, translator_verbose=translator_verbose,
    )
    pdf_path = model.toPDF(
        url=url, path=pdf_path, out_dir=out_dir,
        journal_type=journal_type, crawl_type=crawl_type, gateway=gateway,
        searchpath=searchpath, template=template, 
        delete_html=delete_html, **gateway_params,
    )
    return pdf_path

def translate_text(argv=sys.argv[1:]):
    """Translate from Japanese to English.

    Args:
        query (str)                 : English to be translated. (required)
        -T/--translator (str)       : Translator identifier, string name of a translator. (default= ``"deepl"`` )
        --browser (bool)            : Whether you want to run Chrome with GUI browser. (default= ``False`` )
        --quiet (bool)              : Whether you want to be quiet or not. (default= ``False`` )
        --translator-verbose (bool) : Whether you want to print translator's output or not. (default= ``False`` )

    Note:
        When you run from the command line, execute as follows::
        
        $ gummy-translate "This is a pen."

    Examples:
        >>> $ gummy-translate "This is a pen."
    """

    parser = argparse.ArgumentParser(prog="gummy-translate", add_help=True)
    parser.add_argument("query", type=str, help="English to be translated")
    parser.add_argument("-T", "--translator", type=str, default="deepl",   help="Translator identifier, string name of a translator")
    parser.add_argument("--browser", action="store_true", help="Whether you want to run Chrome with GUI browser.")
    parser.add_argument("--quiet",              action="store_true",  help="Whether you want to be quiet or not. (default=False)")
    parser.add_argument("--translator-verbose", action="store_true",  help="Whether you want to print translator's output or not. (default=False)")
    args = parser.parse_args(argv)

    chrome_options = get_chrome_options(browser=args.browser)
    query = args.query
    translator = args.translator
    verbose = not args.quiet
    translator_verbose = args.translator_verbose

    model = TranslationGummy(
        chrome_options=chrome_options, gateway="useless", translator=translator,
        verbose=verbose, translator_verbose=translator_verbose,
    )
    japanese = model.en2ja(query=query)
    return japanese



#coding: utf-8
""" Utility programs for handling compression and decompression """
import os
import shutil
import bz2
import gzip
import zipfile
import tarfile
import mimetypes
from abc import ABCMeta, abstractstaticmethod

from .coloring_utils import toBLUE, toRED
from .generic_utils import recreate_dir

def get_mimetype_mimetypes(path):
    """Guess the type of a file based on its URL (filename).

    Args:
        path (str) : filename
    
    Returns:
        - ``None`` (if the type can't be guessed (no or unknown suffix)
        -  a string of the form ``type/subtype`` (otherwise)

    Examples:
        >>> get_mimetype_mimetypes("gummy.zip")
        'application/zip'
        >>> get_mimetype_mimetypes("gummy.tar.gz")
        'application/x-tar'
        # Check the difference when `path` does not exist.
        >>> os.path.exists("gummy.png")
        True
        >>> get_mimetype_mimetypes("gummy.png")
        'image/png'
        >>> os.path.exists("gummy_.png")
        False
        >>> get_mimetype_mimetypes("gummy_.png")
        'image/png'
    """
    return mimetypes.guess_type(path)[0]

def get_mimetype_libmagic(path):
    """Accepts a filename and returns the detected filetype.
    
    Args:
        path (str) : filename

    Returns:
        str : 
            - ``None`` (if the file does not exist)
            -  a string of the form ``type/subtype`` (otherwise)
    
    Examples:
        >>> get_mimetype_libmagic("gummy.zip")
        'application/zip'
        >>> get_mimetype_libmagic("gummy.tar.gz")
        'application/x-tar'
        # Check the difference when `path` does not exist.
        >>> os.path.exists("gummy.png")
        True
        >>> get_mimetype_libmagic("gummy.png")
        'image/png'
        >>> os.path.exists("gummy_.png")
        False
        >>> print(get_mimetype_mimetypes("gummy_.png"))
        None
    """
    return magic.from_file(path, mime=True) if os.path.exists(path) else None

try:
    import magic
    get_mimetype = get_mimetype_libmagic
except ImportError:
    print(f"failed to find {toBLUE('libmagic')}, so use {toBLUE('mimetypes')} instead.")
    get_mimetype = get_mimetype_mimetypes

def is_compressed(ext):
    """Check whether file is compressed or not from the extensions."""
    return ext in [".zip", ".gz", ".tar.gz", ".tgz", "bzip2", ".tar.bz2", ".tar"]

def extract_from_compressed(path, ext=None, dirname=".", verbose=True):
    """Extract files from compressed file.

    Args:
        path (str)      : path/to/compressed_file.
        ext (str)       : Extract only files with this extension from compressed files. If ``None``, all files will be extracted.
        dirname (str)   : Where the extracted file will be stored.
        verbose (bool)  : Whether print names in extracted file or not.

    Returns:
        list : Paths of extracted files.
    """
    zip_ext = os.path.splitext(path)[-1]
    if zip_ext == "":
        mimetype = get_mimetype(path)
        zip_ext = ".zip" if (mimetype is not None) and (mimetype.split("/")[-1] == "zip") else None
    Extractor = {
        ".zip" : ZipExtractor
    }.get(zip_ext, TarExtractor)
    extracted_file_paths = Extractor.extract_from_compressed(
        path=path, ext=ext, dirname=dirname, verbose=verbose,
    )
    return extracted_file_paths

class GummyAbstExtractor(metaclass=ABCMeta):
    """File Extractor."""
    @classmethod
    def extract_from_compressed(cls, path, ext=None, dirname=".", verbose=True):
        """Extract files from compressed file.

        Args:
            path (str)      : path/to/compressed_file.
            ext (str)       : Extract only files with this extension from compressed files. If ``None``, all files will be extracted.
            dirname (str)   : Where the extracted file will be stored.
            verbose (bool)  : Whether print names in extracted file or not.

        Returns:
            list : Paths of extracted files.
        """
        extracted_file_paths = []
        print(f"Contents in {toBLUE(path)}:")
        with cls.open_compressed_file(path) as compressed_f:
            for name in cls.get_namelist(compressed_f):
                if ext is None or name.endswith(ext):
                    compressed_f.extract(name, path=dirname)
                    extracted_file_path = os.path.join(dirname, name)
                    extracted_file_paths.append(extracted_file_path)
                    name += f" (Save at {toBLUE(extracted_file_path)})"
                if verbose: print(f"\t- {name}")
        return extracted_file_paths

    @abstractstaticmethod
    def open_compressed_file(path):
        """Open a compressed file."""
        return open(path)

    @abstractstaticmethod
    def get_namelist(compressed_f):
        """Get name list in the extracted file."""
        for name in compressed_f.namelist():
            yield name
            
class ZipExtractor(GummyAbstExtractor):
    """Extractor for Zip file.

    .. code-block:: python

        >>> import zipfile
        >>> with zipfile.ZipFile(path) as f:
        ...     for name in f.get_namelist():
        ...         print(name)
    """
    @staticmethod
    def open_compressed_file(path):
        return zipfile.ZipFile(path)
    
    @staticmethod
    def get_namelist(compressed_f):
        for name in compressed_f.namelist():
            yield name
            
class TarExtractor(GummyAbstExtractor):
    """Extractor for Tar file.
    
    .. code-block:: python

        >>> import tarfile
        >>> with tarfile.open(path) as f:
        ...     for m in f.getmembers():
        ...         name = m.name
        ...         print(name)
    """
    @staticmethod
    def open_compressed_file(path):
        return tarfile.open(path)
    
    @staticmethod
    def get_namelist(compressed_f):
        for m in compressed_f.getmembers():
            name = m.name
            yield name


#coding: utf-8
"""This program file is only for `documents <https://iwasakishuto.github.io/Translation-Gummy/index.html>`_. 
You can see the same contents by looking at `Github <https://github.com/iwasakishuto/Translation-Gummy/tree/master/gummy/templates>`_ .
"""
import os
from gummy.utils import TEMPLATES_DIR

def _mk_func(fn):  
    fp = os.path.join(TEMPLATES_DIR, fn)
    with open(fp, mode="r") as f:
        code = "\t".join(f.readlines())
    def func():
        return fp
    func.__doc__ = f"""Return ``TEMPLATES_DIR``/{fn}
    
    Returns:
        str : Where this file is.

    The contents of the file are as follows:

    .. code-block:: html
        
        {code}
    """
    return func

for fn in os.listdir(TEMPLATES_DIR):
    if fn[-5:] != ".html": continue
    name = fn.replace(".", "_")
    exec(f"{name} = _mk_func('{fn}')")



# coding: utf-8

__copyright__       = "Copyright (C) 2020 Shuto Iwasaki"
__version__         = "3.5.2"

__license__         = "MIT"
__author__          = "Shuto Iwasaki"
__author_twitter__  = "https://twitter.com/cabernet_rock"
__author_email__    = "cabernet.rock@gmail.com"
__url__             = "https://github.com/iwasakishuto/Translation-Gummy"

from . import gateways
from . import journals
from . import models
from . import translators
from .models import TranslationGummy


# coding: utf-8
import os
import setuptools
import gummy

DESCRIPTION = "Translation Gummy is a magical gadget which enables user to be able to speak and understand other languages."

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()
with open("requirements.txt", mode="r") as f:
    INSTALL_REQUIRES = [line.rstrip("\n") for line in f.readlines() if line[0]!=("#")]

def setup_package():
    metadata = dict(
        name="Translation-Gummy",
        version=gummy.__version__,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        author="Shuto Iwasaki",
        author_email="cabernet.rock@gmail.com",
        license="MIT",
        project_urls={
            "Documentation" : "https://iwasakishuto.github.io/Translation-Gummy/index.html",
            "Bug Reports"   : "https://github.com/iwasakishuto/Translation-Gummy/issues",
            "Source Code"   : "https://github.com/iwasakishuto/Translation-Gummy",
            "Say Thanks!"   : "https://twitter.com/cabernet_rock",
        },
        packages=setuptools.find_packages(),
        package_data={"gummy": ["templates/*"]},
        python_requires=">=3.7",
        install_requires=INSTALL_REQUIRES,
        extras_require={
          "tests": ["pytest"],
        },
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Environment :: Console",
            "Intended Audience :: Other Audience",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Topic :: Software Development :: Libraries",
            "Topic :: Software Development :: Libraries :: Python Modules",
        ],
        entry_points = {
            "console_scripts": [
                "gummy-journal=gummy.cli:translate_journal",
                "gummy-translate=gummy.cli:translate_text",
        ],
    },
    )
    setuptools.setup(**metadata)

if __name__ == "__main__":
    setup_package()


#coding: utf-8
import os
from pathlib import Path
from .coloring_utils  import toBLUE

__all__ = [
    "UTILS_DIR", "MODULE_DIR", "TEMPLATES_DIR", "REPO_DIR", "GUMMY_DIR", 
    "DOTENV_PATH",
    "IMG_NOT_FOUND_SRC"
]

UTILS_DIR     = os.path.dirname(os.path.abspath(__file__))      # path/to/gummy/utils
MODULE_DIR    = os.path.dirname(UTILS_DIR)                      # path/to/gummy
TEMPLATES_DIR = os.path.join(MODULE_DIR, "templates")           # path/to/gummy/templates
REPO_DIR      = os.path.dirname(MODULE_DIR)                     # path/to/Translation-Gummy
GUMMY_DIR     = os.path.join(os.path.expanduser('~'), '.gummy') # /Users/<username>/.gummy
# Check whether uid/gid has the write access to DATADIR_BASE
if os.path.exists(GUMMY_DIR) and not os.access(GUMMY_DIR, os.W_OK):
    GUMMY_DIR = os.path.join('/tmp', '.gummy')
if not os.path.exists(GUMMY_DIR):
    os.mkdir(GUMMY_DIR)
    print(f"{toBLUE(GUMMY_DIR)} is created. Downloaded data will be stored here.")
DOTENV_PATH   = os.path.join(GUMMY_DIR, ".env")                 # /Users/<username>/.gummy/.env
if not os.path.exists(DOTENV_PATH):
    Path(DOTENV_PATH).touch()
    print(f"{toBLUE(DOTENV_PATH)} is created. Environment variables should be stored here.")
IMG_NOT_FOUND_SRC = "https://github.com/iwasakishuto/Translation-Gummy/blob/master/image/image-not-found.png?raw=true"



# coding: utf-8
from gummy.utils import get_driver
from gummy import translators

def _test_translators(db, identifier, **kwargs):
    with get_driver() as driver:
        translator = translators.get(identifier, driver=driver, maxsize=1000)
        ja_simple_sentence    = translator.en2ja(query=db.sentences.get("simple_sentence"))
        ja_multiple_sentences = translator.en2ja(query=db.sentences.get("multiple_sentences"))
        ja_SteveJobsSpeech    = translator.en2ja(query=db.sentences.get("SteveJobsSpeech"))

    assert len(ja_simple_sentence) > 0
    assert len(ja_multiple_sentences) > 0
    assert len(ja_SteveJobsSpeech) > 0

def test_google_translators(db):
    _test_translators(db=db, identifier="google")

def test_deepl_translators(db):
    _test_translators(db=db, identifier="deepl")


# coding: utf-8
import os
import pytest
from gummy.models import TranslationGummy
from gummy import gateways
from gummy import translators
from gummy.utils import get_driver

from data import JournalData

@pytest.mark.parametrize("gateway", list(gateways.all.keys()))
@pytest.mark.parametrize("translator", list(translators.all.keys()))
def test_models(db, gateway, translator, journal_type="nature"):
    url = db.journals.get(journal_type)[0]
    with get_driver() as driver:
        gummy = TranslationGummy(driver=driver, gateway=gateway, translator=translator)    
        # Make HTML & PDF.
        pdfpath  = gummy.toPDF(url=url, delete_html=True)
        os.remove(pdfpath)



# coding: utf-8
"""Utility programs for handling and analyzing PDF file."""
import io
import os
import base64
import urllib
import werkzeug
import contextlib
from pdfminer.converter import PDFPageAggregator
from pdfminer.layout import LAParams, LTContainer, LTTextBox, LTImage, LTTextLine, LTFigure
from pdfminer.pdfinterp import PDFPageInterpreter, PDFResourceManager
from pdfminer.pdfpage import PDFPage

from ._path import GUMMY_DIR
from .coloring_utils import toRED, toBLUE
from .download_utils import download_file

@contextlib.contextmanager
def get_pdf_pages(file, dirname=GUMMY_DIR):
    """Get PDF pages.

    Args:
        file (data, str) : url or path or data of PDF.
        dirname (str)    : if ``file`` is url, download and save it to ``dirname``. (defalt= ``GUMMY_DIR``)
    """
    if isinstance(file, werkzeug.datastructures.FileStorage) or isinstance(file, io.TextIOWrapper):
        yield PDFPage.get_pages(fp=file)
    else:
        if isinstance(file, str) and (not os.path.exists(file)):
            path = download_file(url=file, dirname=dirname)
            if path is None:
                print(toRED(f"Failed to download PDF from {toBLUE(file)}"))
        else:
            path = file
        with open(path, mode="rb") as f_pdf:
            yield PDFPage.get_pages(fp=f_pdf)

def parser_pdf_pages(layout_objs):
    """Parse PDF pages and get contents in order.

    Args:
        layout_objs (list) : Each element is pdfminer.layout object.

    Returns:
        texts (list) : Each element is text (str).
    """
    texts = []
    for lt_obj in layout_objs:
        if isinstance(lt_obj, LTTextBox) or isinstance(lt_obj, LTTextLine):
            texts.append(lt_obj.get_text())
        elif isinstance(lt_obj, LTImage):
            rawdata = lt_obj.stream.get_rawdata()
            bs64data = base64.b64encode(rawdata).decode("utf-8")
            texts.append(f'<img src="data:image/jpeg;base64,{bs64data}" />')
        elif isinstance(lt_obj, LTFigure):
            texts.extend(parser_pdf_pages(lt_obj._objs))
    return texts

def get_pdf_contents(file, dirname=GUMMY_DIR):
    """Get PDF contents.

    Args:
        file (data, str) : url or path or data of PDF.
        dirname (str)    : if ``file`` is url, download and save it to ``dirname``. (defalt= ``GUMMY_DIR``)
    """
    # Settings.
    rsrcmgr     = PDFResourceManager()
    laparams    = LAParams(detect_vertical=True)
    device      = PDFPageAggregator(rsrcmgr=rsrcmgr, laparams=laparams)
    interpreter = PDFPageInterpreter(rsrcmgr=rsrcmgr, device=device)
    #  parse PDF pages
    pdf_pages = []
    with get_pdf_pages(file=file, dirname=dirname) as pages:
        for page in pages:
            interpreter.process_page(page)
            layout = device.get_result()
            pdf_pages.append(parser_pdf_pages(layout_objs=layout._objs))
    return pdf_pages
